from jailconf.exceptions import ConfSemanticError
from jailconf.lexer import Lexer

from collections import OrderedDict
import os
import re
from tempfile import mkstemp

class Append(object):
    def __init__(self, param, value):
        self.param = param
        self.value = value

class Conf(OrderedDict):

    valid_param_rgx = re.compile('^%s$' % Lexer.t_NAME.__doc__)
    valid_value_rgx = re.compile('^(%s)$' % '|'.join((
        Lexer.t_NAME.__doc__,
        Lexer.t_SINGLE_QUOTED_STRING.__doc__,
        Lexer.t_DOUBLE_QUOTED_STRING.__doc__
    )))
    
    @classmethod
    def validate_param(cls, param):
        if cls.valid_param_rgx.match(param):
            return param
        else:
            raise ValueError("Invalid parameter %s" % repr(param))
            
    @classmethod
    def validate_value(cls, value):
        if cls.valid_value_rgx.match(value):
            return value
        else:
            raise ValueError("Invalid value %s" % repr(value))
            
    @classmethod
    def validate_key_value(cls, key, value):
        if isinstance(value, JailBlock):
            cls.validate_value(key)
        else:
            cls.validate_param(key)
            if isinstance(value, str):
                cls.validate_value(value)
            elif isinstance(value, list):
                if len(value) < 2:
                    raise ValueError("Lists of values should have at least 2 elements.")
                for v in value:
                    cls.validate_value(v)
            elif value is not True:
                raise ValueError(
                    "Invalid type for value %s. Should be of type str, list, JailBlock, or should be the value True."
                    % repr(value)
                )
        return key, value
         
    def update(self, *args, **kwargs):
        if len(args) > 1:
            raise TypeError("Expected at most 1 argument, got 2")
        if args:
            if hasattr(args[0], 'keys'):
                lst = [(k,args[0][k]) for k in args[0].keys()]
            else:
                lst = args[0]
        else:
            lst = []
        if kwargs:
            lst.extend(kwargs.items())
        for item in lst:
            if isinstance(item, Append):
                if item.param in self:
                    if isinstance(self[item.param], str):
                        super(Conf, self).__setitem__(item.param, [self[item.param], item.value])
                    elif isinstance(self[item.param], list):
                        self[item.param].append(item.value)
                    else:
                        raise ConfSemanticError(
                            "Trying to append to parameter %s but it has type %s" % (
                                item.param, type(self[item.param])
                            )
                        )
                else:
                    raise ConfSemanticError(
                        "Trying to append to undefined parameter %s" % item.param
                    )
            else:
                key, value = item
                self[key] = value

        
    def __setitem__(self, k, v):
        self.__class__.validate_key_value(k, v)
        super(Conf, self).__setitem__(k, v)
    
    def __init__(self, *args, **kwargs):
        super(Conf, self).__init__()
        self.update(*args, **kwargs)
        
    def strgen(self, indentation = '\t', current_indent = 0):
        for key, value in self.items():
            yield current_indent * indentation
            yield key
            if isinstance(value, JailBlock):
                yield ' {\n'
                yield from value.strgen(indentation, current_indent + 1)
                yield '}\n'
            elif value is True:
                yield ';\n'
            elif isinstance(value, list):
                yield ' = %s;\n' % (', '.join(value))
            else:
                yield ' = %s;\n' % value
    
    def dumps(self, indentation = '\t', current_indent = 0):
        return ''.join(self.strgen(indentation, current_indent))
        
    def write(self, path, indentation = '\t'):
        handle, temp_path = mkstemp(prefix = path)
        os.write(handle, self.dumps(indentation = indentation).encode('utf-8'))
        os.close(handle)
        os.rename(temp_path, path)
    
class JailConf(Conf):
    def jails(self):
        for key, value in self.items():
            if isinstance(value, JailBlock):
                yield (key, value)
        
class JailBlock(Conf):
    @classmethod
    def validate_key_value(cls, k, v):
        if isinstance(v, JailBlock):
            raise ValueError("Jail block not allowed in this context.")
        return super(JailBlock, cls).validate_key_value(k, v)



"""
Parse and edit your FreeBSD jail.conf file with python.

Examples:

import jailconf
conf = jailconf.load('/etc/jail.conf')

conf['path'] = '"/var/jail/$name"'
conf['mount.devfs'] = True
conf['myjail'] = jailconf.JailBlock()
conf['myjail']['ip4.addr'] = ['10.0.0.1', '10.0.0.2']
conf['myjail']['host.hostname'] = 'example.com'

conf.write('/etc/jail.conf') # overwrite configuration file

A bit more documentation in the README.rst.
    https://github.com/leforestier/jailconf/blob/master/README.rst
"""

from jailconf.parser import loads, load
from jailconf.structures import JailConf, JailBlock

__author__ = "Benjamin Le Forestier (benjamin@leforestier.org)"
__version__ = '0.2.1'



from jailconf.exceptions import ConfSyntaxError
from jailconf.lexer import Lexer
from jailconf.structures import JailConf, JailBlock, Append

from ply import yacc

__all__ = ['loads']

class Parser(object):

    def p_conf(self, p):
        " conf : global_scope "
        p[0] = p[1]

    def p_global_scope_1(self, p):
        """ global_scope : global_scope statement
            global_scope : global_scope jail_definition """
        p[0] = p[1]
        p[0].update((p[2],))

    def p_global_scope_empty(self, p):
        " global_scope : "
        p[0] = JailConf()

    def p_jail_block(self, p):
        " jail_definition : value OPEN_BRACE statement_list CLOSE_BRACE "
        p[0] = (p[1], JailBlock(p[3]))

    def p_statement_list(self, p):
        " statement_list : statement_list statement "
        p[0] = p[1] + (p[2],)

    def p_statement_list_empty(self, p):
        " statement_list : "
        p[0] = ()

    def p_statement(self, p):
        """ statement : affectation
            statement : declaration
            statement : append """
        p[0] = p[1]

    def p_affectation(self, p):
        """ affectation : parameter EQUAL value SEMICOLON
            affectation : parameter EQUAL many_values SEMICOLON """
        p[0] = (p[1], p[3])

    def p_append(self, p):
        " append : parameter PLUS_EQUAL value SEMICOLON "
        p[0] = Append(p[1], p[3])

    def p_declaration(self, p):
        " declaration : parameter SEMICOLON "
        p[0] = (p[1], True)

    def p_parameter(self, p):
        " parameter : NAME "
        p[0] = p[1]

    def p_value(self, p):
        """ value : NAME
            value : SINGLE_QUOTED_STRING
            value : DOUBLE_QUOTED_STRING """
        p[0] = p[1]

    def p_many_values(self, p):
        " many_values : many_values COMMA value "
        p[0] = p[1]
        p[0].append(p[3])

    def p_many_values_only_two(self, p):
        " many_values : value COMMA value "
        p[0] = [p[1], p[3]]

    def p_error(self, p):
        raise ConfSyntaxError(repr(p))

    def __init__(self, lexer = None):
        lexer = lexer or Lexer()
        self.tokens = lexer.tokens
        self._lexer = lexer
        self._parser = yacc.yacc(module=self, debug=False, write_tables=0)

    def parse(self, entry):
        return self._parser.parse(
            entry,
            lexer = self._lexer.lexer
        )

loads = Parser().parse

def load(path):
    with open(path) as fp:
        return loads(fp.read())

if __name__ == '__main__':
    import sys
    text = sys.stdin.read()
    print()
    print(loads(text).dumps())



from setuptools import setup
import sys

with open('README.rst') as fd:
    long_description = fd.read()

setup(
    name='jailconf',
    version='0.2.1',
    packages=['jailconf'],
    install_requires = ['ply>=3.9'],
    python_requires='>=3.0',
    author = 'Benjamin Le Forestier',
    author_email = 'benjamin@leforestier.org',
    url = 'https://github.com/leforestier/jailconf',
    keywords = ["jail.conf", "parse", "edit", "configuration", "freebsd", "jail"],
    description = "Parse and edit your FreeBSD jail.conf file",
    long_description = long_description,
    classifiers = [
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: System :: Systems Administration',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



from jailconf.exceptions import ConfTokenError

from ply import lex
import re


class Lexer(object):

    tokens = (
        'OPEN_BRACE',
        'CLOSE_BRACE',
        'EQUAL',
        'COMMA',
        'PLUS_EQUAL',
        'SEMICOLON',
        'NAME',
        'DOUBLE_QUOTED_STRING',
        'SINGLE_QUOTED_STRING'
    )

    t_OPEN_BRACE = '{'
    t_CLOSE_BRACE = '}'
    t_EQUAL = '='
    t_COMMA = ','
    t_PLUS_EQUAL = r'\+='
    t_SEMICOLON = ';'

    def t_ignore_NEW_LINE(self, token):
        r'\n'

    def t_ignore_C_STYLE_COMMENT(self, token):
        r'/\*((?!\*/).|\n)*\*/'

    def t_ignore_CPP_STYLE_COMMENT(self, token):
        r'//.*'

    def t_ignore_SHELL_STYLE_COMMENT(self, token):
        r'\#.*'

    def t_NAME(self, token):
        r'\$?[-a-zA-Z0-9._/*:]+'
        return token

    def t_DOUBLE_QUOTED_STRING(self, token):
        r'"(\\\\|\\"|[^"\\])*"'
        return token

    def t_SINGLE_QUOTED_STRING(self, token):
        r"'(\\\\|\\'|[^'\\])*'"
        return token

    t_ignore_space = r'[ \t]+'

    def t_error(self, token):
        raise ConfTokenError(
            "Illegal character '%s', line %s" % (
                token.value[0],
                token.lineno
            )
        )

    def __init__(self):
        self.lexer = lex.lex(object = self, debug = 0,reflags = re.MULTILINE)

if __name__ == '__main__':
    import sys
    lexer = Lexer().lexer
    lexer.input(sys.stdin.read())
    print()
    while True:
        tok = lexer.token()
        if not tok:
            break      # No more input
        print(tok)
